"""
:summary: rundeckrun package

:license: Creative Commons Attribution-ShareAlike 3.0 Unported
:author: Mark LaPerriere
:contact: rundeckrun@mindmind.com
:copyright: Mark LaPerriere 2013
"""
__docformat__ = "restructuredtext en"

__VERSION__ = (0, 1, 12)
VERSION = '.'.join(map(str, __VERSION__))
